@@ -1,4 +1,4 @@
-/* cups-pdf.c -- CUPS Backend (version 2.5.1, 2011-02-18)
+/* cups-pdf.c -- CUPS Backend (version 2.6.0, 2011-10-01)
    08.02.2003, Volker C. Behr
    Experimentelle Physik V, Universitaet Wuerzburg 
    behr@physik.uni-wuerzburg.de
@@ -423,6 +423,34 @@ static int preparetitle(char *title) {
   return strcmp(title, "");
 }
 
+static char *fgets2(char *fbuffer, int fbufsize, FILE *ffpsrc) {
+  /* like fgets() but linedelimiters are 0x0A, 0x0C, 0x0D (LF, FF, CR). */
+  int c, pos;
+  char *result;
+
+  if (!conf.fixnewlines)
+    return fgets(fbuffer, fbufsize, ffpsrc);
+
+  result=NULL;
+  pos=0;
+
+  while (pos < fbufsize) {      /* pos in [0..fbufsize-1] */
+    c=fgetc(ffpsrc);          /* converts CR/LF to LF in some OSses */
+    if (c == EOF)           /* EOF _or_ error */
+      break;
+    fbuffer[pos++]=c;
+    if (c == 0x0A || c == 0x0C || c == 0x0D) /* line is at an end */
+      break;
+  }
+
+  if (pos > 0 && !ferror(ffpsrc)) { /* at least one char read and no error */
+    fbuffer[pos]='\0';
+    result=fbuffer;
+  }
+
+  return result;
+}
+
 static int preparespoolfile(FILE *fpsrc, char *spoolfile, char *title, char *cmdtitle, 
                      int job, struct passwd *passwd) {
   cp_string buffer;
@@ -447,7 +475,11 @@ static int preparespoolfile(FILE *fpsrc, char *spoolfile, char *title, char *cmd
   }
   log_event(CPDEBUG, "owner set for spoolfile", spoolfile);
   rec_depth=0;
-  while (fgets(buffer, BUFSIZE, fpsrc) != NULL) {
+  if (conf.fixnewlines)
+    log_event(CPSTATUS, "***Experimental Option: FixNewlines", NULL);
+  else
+    log_event(CPDEBUG, "using traditional fgets", NULL);
+  while (fgets2(buffer, BUFSIZE, fpsrc) != NULL) {
     if (!strncmp(buffer, "%!", 2)) {
       log_event(CPDEBUG, "found beginning of postscript code", buffer);
       break;
@@ -455,7 +487,7 @@ static int preparespoolfile(FILE *fpsrc, char *spoolfile, char *title, char *cmd
   }
   log_event(CPDEBUG, "now extracting postscript code", NULL);
   (void) fputs(buffer, fpdest);
-  while (fgets(buffer, BUFSIZE, fpsrc) != NULL) {
+  while (fgets2(buffer, BUFSIZE, fpsrc) != NULL) {
     (void) fputs(buffer, fpdest);
     if (!is_title && !rec_depth)
       if (sscanf(buffer, "%%%%Title: %"TBUFSIZE"c", title)==1) {
@@ -535,6 +567,8 @@ int main(int argc, char *argv[]) {
   int size;
   mode_t mode;
   struct passwd *passwd;
+  gid_t *groups;
+  int ngroups;
   pid_t pid;
 
   if (setuid(0)) {
@@ -624,6 +658,29 @@ int main(int argc, char *argv[]) {
     log_event(CPDEBUG, "output directory name generated", dirname);
     mode=(mode_t)(0666&~conf.userumask);
   }
+  ngroups=32;
+  groups=calloc(ngroups, sizeof(gid_t));
+  if (groups == NULL) {
+    (void) fputs("CUPS-PDF: failed to allocate memory\n", stderr);
+    free(user);
+    if (logfp!=NULL)
+      (void) fclose(logfp);
+    return 5;
+  }
+  size=getgrouplist(user, passwd->pw_gid, groups, &ngroups);
+  if (size == -1) {
+    free(groups);
+    groups=calloc(ngroups, sizeof(gid_t));
+    size=getgrouplist(user, passwd->pw_gid, groups, &ngroups);
+  }
+  if (size < 0) {
+    log_event(CPERROR, "getgrouplist failed", NULL);
+    free(user);
+    free(groups);
+    if (logfp!=NULL)
+      (void) fclose(logfp);
+    return 5;
+  }
   free(user);
   if (prepareuser(passwd, dirname)) {
     free(dirname);
@@ -721,11 +778,15 @@ int main(int argc, char *argv[]) {
 
     if (setgid(passwd->pw_gid))
       log_event(CPERROR, "failed to set GID for current user", NULL);
-    else 
+    else
       log_event(CPDEBUG, "GID set for current user", NULL);
+    if (setgroups(ngroups, groups))
+      log_event(CPERROR, "failed to set supplementary groups for current user", NULL);
+    else 
+      log_event(CPDEBUG, "supplementary groups set for current user", NULL);
     if (setuid(passwd->pw_uid))
       log_event(CPERROR, "failed to set UID for current user", passwd->pw_name);
-    else 
+    else
       log_event(CPDEBUG, "UID set for current user", passwd->pw_name);
      
     (void) umask(0077);
